
/******************************************************************************
* MODULE     : rational.hpp
* DESCRIPTION: Rational numbers
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_RATIONAL_HPP
#define __MML_RATIONAL_HPP
#include "integer.hpp"
namespace mml {

class rational;
class rational_rep;

class rational_rep REP_STRUCT {
private:
  mpq_t x;
public:
  inline rational_rep () { mpq_init (x); }
  inline ~rational_rep () { mpq_clear (x); }
  friend class rational;
};

class rational {
INDIRECT_PROTO (rational, rational_rep)
private:
  inline mpq_t& operator * () const { return rep->x; }

public:
  // constructors
  inline rational (): rep (new rational_rep ()) {}
  inline rational (signed int i): rep (new rational_rep ()) {
    mpq_set_si (rep->x, i, 1); }
  inline rational (unsigned int i):
    rep (new rational_rep ()) { mpq_set_ui (rep->x, i, 1); }
  inline rational (signed short int i):
    rep (new rational_rep ()) { mpq_set_si (rep->x, i, 1); }
  inline rational (unsigned short int i):
    rep (new rational_rep ()) { mpq_set_ui (rep->x, i, 1); }
  inline rational (signed long int i):
    rep (new rational_rep ()) { mpq_set_si (rep->x, i, 1); }
  inline rational (unsigned long int i):
    rep (new rational_rep ()) { mpq_set_ui (rep->x, i, 1); }
  inline rational (const integer& i):
    rep (new rational_rep ()) { mpq_set_z (rep->x, *i); }

  // getting the numerator and denominator
  inline friend integer num (const rational& x1) {
    integer r; mpq_get_num (*r, *x1); return r; }
  inline friend integer den (const rational& x1) {
    integer r; mpq_get_den (*r, *x1); return r; }

  // input/output
  inline friend ostream& operator << (ostream& out, const rational& x) {
    return out << as_generic (x); }

  // basic arithmetic
  inline friend rational operator - (const rational& x1) {
    rational r; mpq_neg (*r, *x1); return r; }
  inline friend rational operator + (const rational& x1, const rational& x2) {
    rational r; mpq_add (*r, *x1, *x2); return r; }
  inline friend rational operator - (const rational& x1, const rational& x2) {
    rational r; mpq_sub (*r, *x1, *x2); return r; }
  inline friend rational operator * (const rational& x1, const rational& x2) {
    rational r; mpq_mul (*r, *x1, *x2); return r; }
  inline friend rational operator / (const rational& x1, const rational& x2) {
    if (mpq_sgn (*x2) == 0) return 0;
    rational r; mpq_div (*r, *x1, *x2); return r; }
  inline friend rational operator % (const rational& x1, const rational& x2) {
    if (mpq_sgn (*x2) == 0) return x1; else return 0; }
  inline friend rational operator << (const rational& x1, long int shift) {
    rational r; mpq_mul_2si (*r, *x1, shift); return r; }
  inline friend rational operator >> (const rational& x1, long int shift) {
    rational r; mpq_mul_2si (*r, *x1, -shift); return r; }
  inline friend rational& operator <<= (rational& x1, long int shift) {
    return x1= x1 << shift; }
  inline friend rational& operator >>= (rational& x1, long int shift) {
    return x1= x1 >> shift; }
  inline friend rational abs (const rational& x1) {
    rational r; mpq_abs (*r, *x1); return r; }

  // comparing rational numbers
  inline friend bool operator == (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) == 0; }
  inline friend bool operator != (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) != 0; }
  inline friend bool operator < (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) < 0; }
  inline friend bool operator > (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) > 0; }
  inline friend bool operator <= (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) <= 0; }
  inline friend bool operator >= (const rational& x1, const rational& x2) {
    return mpq_cmp (*x1, *x2) >= 0; }
  inline friend int sign (const rational& x1) {
    int r= mpq_sgn (*x1); return (r<0? -1: (r==0? 0: 1)); }
  inline friend rational min (const rational& x1, const rational& x2) {
    return x1 <= x2? x1: x2; }
  inline friend rational max (const rational& x1, const rational& x2) {
    return x1 >= x2? x1: x2; }

  // miscellaneous functions
  inline friend nat hash (const rational& x) {
    __mpq_struct &rep= (*x)[0];
    if (rep._mp_num._mp_size == 0) return 0;
    return
      ((nat) (rep._mp_num._mp_d[0])) ^
      ((nat) (rep._mp_num._mp_size << 3)) ^
      ((nat) (rep._mp_den._mp_d[0] << 7)) ^
      ((nat) (rep._mp_den._mp_size << 11));
  }
};
INDIRECT_IMPL (rational, rational_rep)

}
#endif // __MML_RATIONAL_HPP
