
/******************************************************************************
* MODULE     : real_div.cpp
* DESCRIPTION: Division of real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "real.hpp"
namespace mml {

class div_real_rep: public real_rep {
private:
  real x, y;
public:
  inline div_real_rep (real x2, real y2):
    real_rep (weight (x2, y2)), x (x2), y (y2) {}
  generic expression () const {
    return over (as_generic (x), as_generic (y)); }
  dyadic get_bound () { return 2 * div_up (bound (x), bound (y)); }
  interval approx (const dyadic& err);
};

interval
div_real_rep::approx (const dyadic& err) {
  if (wider (best, err)) {
    if (has_huge_prec (best)) {
      dyadic rel_r= err / magnitude (best);
      dyadic rel_x= spread (rel_r, x->wt, wt);
      dyadic rel_y= spread (rel_r, y->wt, wt);
      interval xa= approximate (x, magnitude (x->best) * rel_x);
      interval ya= approximate (y, magnitude (y->best) * rel_y);
      mp_prec_t new_prec= mult_algo (err);
      best= reduce (xa, new_prec) / reduce (ya, new_prec);
    }
    else {
      interval xa= approximate (x, lp_infinity);
      if (is_epsilon (xa)) return xa;
      dyadic bound_x= bound (x);
      dyadic bound_y= bound (y);
      dyadic bound_r= bound_x / bound_y;
      dyadic rel_r  = min (err / bound_r, 1);
      if ((rel_r << (DEFAULT_LOW<<1)) < 1) {
	// Since we lack information for making good a priori error estimates,
	// we first compute an approximation at a more modest precision and
	// then check whether we may use the huge precision error estimates.
	dyadic modest_err= sqrt (rel_r) * bound_r;
	(void) approx (modest_err);
	if (has_huge_prec (best)) return approx (err);
      }
      dyadic rel_x= spread (rel_r, x->wt, wt);
      dyadic rel_y= spread (rel_r, y->wt, wt);
      xa= approximate (x, rel_x * bound_x);
      interval ya= approximate (y, rel_y * bound_y);
      mp_prec_t new_prec= mult_algo (err);
      best= reduce (xa, new_prec) / reduce (ya, new_prec);
    }
    ENSURE_REAL_CONSISTENCY (err);
  }
  return best;
}

real
operator / (const real& x, const real& y) {
  if (normalize (y, MAX_CANCEL_TEST, error (1))) return 0;
  return real (new div_real_rep (x, y));
}

real
operator % (const real& x, const real& y) {
  if (normalize (y, MAX_CANCEL_TEST, error (1))) return x;
  return 0;
}

} // namespace mml
