
/******************************************************************************
* MODULE     : jet.hpp
* DESCRIPTION: low level operations on polynomial or serial segments
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__JET__HPP
#define __MML__JET__HPP
#include "segment.hpp"

namespace mml {
#define TMPL template<class C>

/******************************************************************************
* Vectorial operations
******************************************************************************/

#define jet_set segment_set
#define jet_clear segment_clear
#define jet_copy segment_copy
#define jet_add segment_add
#define jet_sub segment_sub
#define jet_neg segment_neg
#define jet_mul_add segment_mul_add

TMPL void jet_mul (C* dest, const C* s, const C& c, nat n) {
  segment_mul (dest, s, c, n); }
TMPL void jet_mul (C* dest, const C& c, nat n) {
  segment_mul (dest, c, n); }
TMPL void jet_div (C* dest, const C* s, const C& c, nat n) {
  segment_div (dest, s, c, n); }
TMPL void jet_div (C* dest, const C& c, nat n) {
  segment_div (dest, c, n); }
TMPL void jet_mod (C* dest, const C* s, const C& c, nat n) {
  segment_mod (dest, s, c, n); }
TMPL void jet_mod (C* dest, const C& c, nat n) {
  segment_mod (dest, c, n); }

TMPL int
jet_sign (C* s, nat n) {
  while (n != 0) {
    int sgn= sign (*s);
    if (sgn != 0) return sgn;
    s++; n--;
  }
  return 0;
}

TMPL int
jet_cmp (const C* s1, const C* s2, nat n) {
  while (n != 0) {
    int sgn= compare (*s1, *s2);
    if (sgn != 0) return sgn;
    s1++; s2++; n--;
  }
  return 0;
}

/******************************************************************************
* Other linear operators
******************************************************************************/

TMPL void
jet_derive (C* dest, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * dest[i];
}

TMPL void
jet_derive (C* dest, const C* s, nat n) {
  for (nat i=1; i<n; i++)
    dest[i-1]= C(i) * s[i];
}

TMPL void
jet_xderive (C* dest, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * dest[i];
}

TMPL void
jet_xderive (C* dest, const C* s, nat n) {
  for (nat i=0; i<n; i++)
    dest[i]= C(i) * s[i];
}

TMPL void
jet_integrate (C* dest, const C* s, nat n) {
  dest[0]= 0;
  for (nat i=0; i<n; i++)
    dest[i+1]= s[i] / C(i+1);
}

TMPL void
jet_q_difference (C* dest, const C* s, const C& q, nat n) {
  C p (1);
  while (n != 0) {
    *dest= p * (*s);
    dest++; s++; n--; p *= q;
  }
}

TMPL void
jet_dilate (C* dest, const C* s, nat p, nat n) {
  jet_clear (dest, (n-1)*p + 1);
  while (n != 0) {
    *dest= *s;
    dest += p; s++; n--;
  }
}

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename C, typename V>
class multiplier {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    jet_clear (dest, n1+n2-1);
    while (n1 != 0) {
      jet_mul_add (dest, *s1, s2, n2);
      dest++; s1++; n1--;
    }
  }
};

template<typename C, typename V> void
jet_mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  multiplier<C,V>::mul (dest, s1, s2, n1, n2);
}

template<typename C, typename V>
class tmultiplier {
public:
  static void tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    // transposed multiplication (or middle product)
    // s1 has length n1, dest has length n2 and s2 has length n1+n2-1
    s2 += n1;
    jet_clear (dest, n2);
    while (n1 != 0) {
      jet_mul_add (dest, *s1, s2, n2);
      s1++; s2--; n1--;
    }
  }
};

template<typename C, typename V> void
jet_tmul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  tmultiplier<C,V>::tmul (dest, s1, s2, n1, n2);
}

/******************************************************************************
* Division
******************************************************************************/

template<typename C, typename V>
class divisor {
public:
  static void div_mod (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    while (n1 >= n2 && n1 != 0) {
      int d= n1-n2;
      C q= s1[n1-1] / s2[n2-1];
      dest[d]= q;
      for (nat i=0; i<n2; i++)
	s1[i+d] -= q * s2[i];
      n1--;
    }
  }
};

template<typename C, typename V> void
jet_div_mod (C* dest, C* s1, const C* s2, nat n1, nat n2) {
  // (s1, n1) contains the numerator on input and the remainder on output
  // (s2, n2) contains the denominator. We assume n2>0 and s2[n2-1] != 0
  // (dest, n1-n2+1) contains the quotient
  divisor<C,V>::div_mod (dest, s1, s2, n1, n2);
}

template<typename C, typename V> void
jet_dicho_div_mod (C* dest, C* s1, const C* s2, nat n1, nat n2, nat Bound) {
  if (n1 < n2);
  else if (n1 < Bound)
    jet_div_mod<C,naive> (dest, s1, s2, n1, n2);
  else {
    nat l= n1-n2+1;
    if (n2 > l) {
      nat d= n2-l;
      jet_dicho_div_mod<C,V> (dest, s1+d, s2+d, n1-d, l, Bound);
      INLINE_ARRAY (C, buf, l+d-1);
      jet_mul<C,V> (buf, dest, s2, l, d);
      jet_sub (s1, buf, l+d-1);
      DELETE_ARRAY (C, buf, l+d-1);      
    }
    else {
      nat h= (n2+1) >> 1;
      nat d= l-h;
      jet_dicho_div_mod<C,V> (dest+d, s1+d, s2, n1-d, n2, Bound);
      jet_dicho_div_mod<C,V> (dest, s1, s2, n1-h, n2, Bound);
    }
  }
}

/******************************************************************************
* Composition
******************************************************************************/

template<typename C, typename V> void
jet_dicho_compose (C* dest, C* p, const C* s1, const C* s2, nat n1, nat n2) {
  // computes compose ((s1, n1), (s2, n2)) and pow ((s2, n2), n1)
  if (n1 == 1) {
    *dest= *s1;
    jet_copy (p, s2, n2);
  }
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2, k2= l2 + d2;
    INLINE_ARRAY (C, c1, l1 + l2 + k1 + k2);
    C* c2= c1 + l1;
    C* p1= c2 + l2;
    C* p2= p1 + k1;
    jet_dicho_compose<C,V> (c1, p1, s1, s2, h1, n2);
    jet_dicho_compose<C,V> (c2, p2, s1+h1, s2, h2, n2);
    jet_mul<C,V> (dest, c2, p1, l2, k1);
    jet_add (dest, c1, l1);
    jet_mul<C,V> (p, p1, p2, k1, k2);
    DELETE_ARRAY (C, c1, l1 + l2 + k1 + k2);
  }
}

template<typename C, typename V> void
jet_dicho_compose (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
  assert (n1 > 0 && n2 > 0);
  if (n1 == 1) *dest= *s1;
  else {
    nat d2= n2 - 1;
    nat h1= n1 >> 1, h2= n1 - h1;
    nat l1= (h1-1) * d2 + 1, l2= (h2-1) * d2 + 1;
    nat k1= l1 + d2;
    INLINE_ARRAY (C, c1, l1 + l2 + k1);
    C* c2= c1 + l1;
    C* p1= c2 + l2;
    jet_dicho_compose<C,V> (c1, p1, s1, s2, h1, n2);
    jet_dicho_compose<C,V> (c2, s1+h1, s2, h2, n2);
    jet_mul<C,V> (dest, c2, p1, l2, k1);
    jet_add (dest, c1, l1);
    DELETE_ARRAY (C, c1, l1 + l2 + k1);
  }
}

#undef TMPL
} // namespace mml
#endif //__MML__JET__HPP
