
/******************************************************************************
* MODULE     : tft_tree.hpp
* DESCRIPTION: main data structure for TFT multiplication
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__TFT_TREE__HPP
#define __MML__TFT_TREE__HPP
#include "segment.hpp"
#include "fft.hpp"

namespace mml {
#define TMPL template<class C>
#define TMPL_DEF template<typename C>
#define Tft_tree tft_tree<C>
#define Tft_tree_rep tft_tree_rep<C>
#define Tft_leaf tft_leaf<C>
#define Tft_binary tft_binary<C>

TMPL class tft_tree_rep;
TMPL class tft_tree;
TMPL Tft_tree& car (Tft_tree& t);
TMPL Tft_tree& cdr (Tft_tree& t);

/******************************************************************************
* Status of Tft tree nodes
******************************************************************************/

struct tft_status {
  // for direct and inverse TFT
  unsigned depth    : 8; // depth = log_2 n, where n is the size of the array
  unsigned is_zero  : 1; // does this node contain zero?
  unsigned is_target: 1; // is this a target range of the tft?
  unsigned cruft    : 6;

  // for inverse TFT only
  unsigned step     : 8; // at which step are we?
  unsigned forward  : 1; // are we going forwards?
  unsigned backward : 1; // are we going backwards?
  unsigned rubbish  : 6;

  inline tft_status (nat depth2):
    depth (depth2), is_zero (true), is_target (false),
    step (0), forward (true), backward (false) {}
  inline ~tft_status () {}
  // friend ostream& operator << (ostream& out, tft_status st);
};

/******************************************************************************
* Tft_tree class and its representation class
******************************************************************************/

TMPL_DEF
struct tft_tree_rep REP_STRUCT {
public:
  tft_status status;

  Tft_tree_rep (nat depth): status (depth) {}
  virtual ~Tft_tree_rep () {}
  virtual bool is_atomic () const = 0;
  virtual bool is_compound () const = 0;
  virtual C*        seg () const { assert (false); exit (1); }
  virtual Tft_tree  car () const { assert (false); exit (1); }
  virtual Tft_tree  cdr () const { assert (false); exit (1); }
  virtual Tft_tree& access_car () { assert (false); exit (1); }
  virtual Tft_tree& access_cdr () { assert (false); exit (1); }
  friend class Tft_tree;
};

TMPL_DEF
class tft_tree {
INDIRECT_PROTO_1 (tft_tree, tft_tree_rep, C)
public:
  friend Tft_tree& car LESSGTR (Tft_tree& t);
  friend Tft_tree& cdr LESSGTR (Tft_tree& t);
};
INDIRECT_IMPL_1 (tft_tree, tft_tree_rep, typename C, C)

TMPL bool is_atomic   (const Tft_tree& t) { return t->is_atomic (); }
TMPL bool is_compound (const Tft_tree& t) { return t->is_compound (); }
TMPL bool is_zero     (const Tft_tree& t) { return t->status.is_zero; }
TMPL C*        seg (const Tft_tree& t) { return t->seg (); }
TMPL Tft_tree  car (const Tft_tree& t) { return t->car (); }
TMPL Tft_tree  cdr (const Tft_tree& t) { return t->cdr (); }
TMPL Tft_tree& car (Tft_tree& t) { return t.rep->access_car (); }
TMPL Tft_tree& cdr (Tft_tree& t) { return t.rep->access_cdr (); }
TMPL nat       D (const Tft_tree& t) { return t->status.depth; }
TMPL nat       N (const Tft_tree& t) { return 1 << t->status.depth; }

/******************************************************************************
* Leafs and binary nodes
******************************************************************************/

TMPL_DEF
class tft_leaf_rep: public Tft_tree_rep {
  C* s;

public:
  tft_leaf_rep (nat depth): Tft_tree_rep (depth), s (NULL) {}
  tft_leaf_rep (nat depth, const C* s2):
    Tft_tree_rep (depth), s (new C[1<<depth]) {
      this->status.is_zero  = false;
      if (s2 == NULL) segment_clear (s, 1<<depth);
      else segment_copy (s, s2, 1<<depth); }
  tft_leaf_rep (nat depth, const C* l, const C* r):
    Tft_tree_rep (depth), s (new C[1<<depth]) {
      nat h= 1 << (depth-1);
      this->status.is_zero  = false;
      segment_copy (s, l, h);
      segment_copy (s+h, r, h); }
  ~tft_leaf_rep () { if (s != NULL) delete[] s; }
  bool is_atomic () const { return true; }
  bool is_compound () const { return false; }
  C* seg () const { return s; }
};

TMPL Tft_tree
tft_leaf (nat depth) {
  return new tft_leaf_rep<C> (depth);
}

TMPL Tft_tree
tft_leaf (nat depth, const C* s) {
  return new tft_leaf_rep<C> (depth, s);
}

TMPL Tft_tree
tft_leaf (nat depth, const C* l, const C* r) {
  return new tft_leaf_rep<C> (depth, l, r);
}

TMPL_DEF
class tft_binary_rep: public Tft_tree_rep {
  Tft_tree l, r;

public:
  tft_binary_rep (const Tft_tree& l2, const Tft_tree& r2):
    Tft_tree_rep (l2->status.depth+1), l (l2), r (r2) {
      this->status.is_zero  = false;
      assert (l->status.depth == r->status.depth); }
  bool is_atomic () const { return false; }
  bool is_compound () const { return true; }
  Tft_tree  car () const { return l; }
  Tft_tree  cdr () const { return r; }
  Tft_tree& access_car () { return l; }
  Tft_tree& access_cdr () { return r; }
};

TMPL Tft_tree
tft_binary (const Tft_tree& l, const Tft_tree& r) {
  return new tft_binary_rep<C> (l, r);
}

/******************************************************************************
* Basic manipulation of Tft_trees
******************************************************************************/

TMPL void
split (Tft_tree& t) {
  assert (is_atomic (t));
  nat depth= t->status.depth - 1;
  if (is_zero (t)) {
    bool is_target= t->status.is_target;
    Tft_tree l= Tft_leaf (depth);
    Tft_tree r= Tft_leaf (depth);
    t= Tft_binary (l, r);
    t->status.is_target= r->status.is_target= t->status.is_target= is_target;
  }
  else {
    bool is_target= t->status.is_target;
    Tft_tree l= Tft_leaf (depth, seg (t));
    Tft_tree r= Tft_leaf (depth, seg (t) + (1<<depth));
    t= Tft_binary (l, r);
    t->status.is_target= r->status.is_target= t->status.is_target= is_target;
  }
}

TMPL void
expand_zero (Tft_tree& t) {
  assert (is_zero (t));
  bool is_target= t->status.is_target;
  t= Tft_leaf (D(t), NULL);
  t->status.is_target= is_target;
}

TMPL void
insert (Tft_tree& t, nat pos, const C* s, nat len) {
  nat n= 1 << t->status.depth;
  assert (pos >= 0 && pos+len <= n);
  if (pos == 0 && len == n)
    t= Tft_leaf (t->status.depth, s);
  else {
    if (is_atomic (t)) split (t);
    if (pos >= (n>>1)) insert (cdr (t), pos-(n>>1), s, len);
    else if (pos+len <= (n>>1)) insert (car (t), pos, s, len);
    else {
      nat half= (n>>1) - pos;
      insert (car (t), pos, s, half);
      insert (cdr (t), 0, s+ half, len - half);
    }
  }
  t->status.is_target= true;
}

TMPL void
simplify (Tft_tree& t) {
  if (is_atomic (t)) return;
  Tft_tree& l= car (t);
  Tft_tree& r= cdr (t);
  simplify (l);
  simplify (r);
  if (is_atomic (l) && is_atomic (r)) {
    if (is_zero (l) && is_zero (r)) {
      bool is_target= l->status.is_target;
      if (r->status.is_target != is_target) return;
      t= Tft_leaf (D(t), NULL);
      t->status.is_target= is_target;
    }
    if (!is_zero (l) && !is_zero (r)) {
      bool is_target= l->status.is_target;
      if (r->status.is_target != is_target) return;
      t= Tft_leaf (D(t), seg (l), seg (r));
      t->status.is_target= is_target;
    }
  }
}

/******************************************************************************
* Displaying Tft_trees and other debugging routines
******************************************************************************/

TMPL void
print (const Tft_tree& t, nat tab= 0) {
  for (nat i=0; i<tab; i++) cout << "  ";
  if (is_atomic (t)) {
    if (is_zero (t)) cout << "Zero [" << N(t) << "]\n";
    else print (seg (t), N(t));
  }
  else {
    cout << "Tree\n";
    print (car (t), tab+1);
    print (cdr (t), tab+1);
  }
}

TMPL Tft_tree
tft_segment (C* s, nat n) {
  Tft_tree t= Tft_leaf (log_2 (n));
  nat start=0, end;
  while (true) {
    while (start < n && s[start] == C(0)) start++;
    end= start;
    if (start == n) break;
    while (end < n && s[end] != C(0)) end++;
    insert (t, start, s+start, end-start);
    start= end;
  }
  return t;
}

TMPL nat
size (const Tft_tree& t) {
  if (is_zero (t)) return 0;
  else if (is_atomic (t)) return N(t);
  else return size (car (t)) + size (cdr (t));
}

#undef TMPL_DEF
#undef TMPL
#undef Tft_tree
#undef Tft_tree_rep
#undef Tft_leaf
#undef Tft_binary
} // namespace mml
#endif //__MML__TFT__HPP
