
/******************************************************************************
* MODULE     : skew_extra.hpp
* DESCRIPTION: skew arithmetic for more general parameterized operators
*              template mechanism does not work with older versions of g++
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__SKEW_EXTRA__HPP
#define __MML__SKEW_EXTRA__HPP
#include "skew_polynomial.hpp"

namespace mml {

/******************************************************************************
* The derivation P |-> A P'
******************************************************************************/

template<typename C, const C& A> class derivation_operator {};

template<typename C, const C& A>
class multiplier<C,skew<derivation_operator<C,A> > > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n1 == 0 || n2 == 0) return;
    INLINE_ARRAY (C, buf, (n1+n2)<<1);
    C* t1 = buf;
    C* t2 = t1 + n1;
    C* aux= t2 + n2;
    jet_clear (dest, n1+n2-1);
    jet_copy (t1, s1, n1);
    jet_copy (t2, s2, n2);
    int i=0;
    C p (1);
    while (n1 > 0) {
      jet_mul<C,karatsuba<2> > (aux, t1, t2, n1, n2);
      jet_div (aux, p, n1+n2-1);
      jet_add (dest, aux, n1+n2-1);
      jet_derive (t1, n1);
      jet_coeff_derive (t2, n2);
      jet_mul (t2, A, n2);
      n1--; i++; p *= C(i);
    }
    DELETE_ARRAY (C, buf, (n1+n2)<<1);
  }
};

template<typename C, const C& A>
class adjoint_helper<C,derivation_operator<C,A> > {
public:
  static void indeterminate (C* dest) {
    dest[0]= -derive (A);
    dest[1]= C(-1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    (void) adj;
    jet_mul<C,skew<derivation_operator<C,A> > > (dest, s1, s2, n1, n2);
  }
};

template<typename C, const C& A> generic
var (const polynomial<C,dense<skew<derivation_operator<C,A> > > >& P) {
  (void) P;
  return times (as_generic (A), MML_PARTIAL);
}

template<typename C, const C& A>
polynomial<C,dense<skew<derivation_operator<C,A> > > >
adjoint (const polynomial<C,dense<skew<derivation_operator<C,A> > > >& P) {
  nat n= dim (P);
  if (n <= 1) return P;
  C* r= new C[n];
  jet_dicho_adjoint<C,derivation_operator<C,A> > (r, seg (P), n, false);
  return polynomial<C,dense<skew<derivation_operator<C,A> > > > (r, n);
}

/******************************************************************************
* The shift operator x |-> x + sh
******************************************************************************/

template<typename K, const K& sh> class shift_operator {};

template<typename K, const K& sh>
class opp_shift {
public:
  static K val;
};

template<typename K, const K& sh> K opp_shift<K,sh>::val= -sh;

template<typename C, typename K, const K& sh>
class multiplier<C,skew<shift_operator<K,sh> > > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_S_operator (dest, s1, s2, sh, n1, n2);
  }
};

template<typename C, typename K, const K& sh>
class adjoint_helper<C,shift_operator<K,sh> > {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(0);
    dest[1]= C(1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    if (adj) jet_mul<C,skew<shift_operator<K,opp_shift<K,sh>::val> > >
	       (dest, s1, s2, n1, n2);
    else jet_mul<C,skew<shift_operator<K,sh> > > (dest, s1, s2, n1, n2);
  }
};

template<typename C, typename K, const K& sh> generic
var (const polynomial<C,dense<skew<shift_operator<K,sh> > > >& P) {
  (void) P;
  return as_generic (MML_ACCESS) (generic ("S"), as_generic (sh));
}

template<typename C, typename K, const K& sh>
polynomial<C,dense<skew<shift_operator<K,opp_shift<K,sh>::val> > > >
adjoint (const polynomial<C,dense<skew<shift_operator<K,sh> > > >& P) {
  nat n= dim (P);
  C* r= new C[n];
  if (n>1) jet_dicho_adjoint<C,shift_operator<K,sh> > (r, seg (P), n, false);
  return polynomial<C,dense<skew<shift_operator<K,opp_shift<K,sh>::val> > > >
    (r, n);
}

/******************************************************************************
* The q-difference operator x |-> q x
******************************************************************************/

template<typename K, const K& q> class q_operator {};

template<typename K, const K& q>
class inv_q {
public:
  static K val;
};

template<typename K, const K& q> K inv_q<K,q>::val= K(1)/q;

template<typename C, typename K> void
jet_coeff_q_difference (C* dest, const K& q, nat n) {
  while (n != 0) {
    *dest= q_difference (*dest, q);
    dest++; n--;
  }
}

template<typename C, typename K> void
mul_q_operator (C* dest, const C* s1, const C* s2,
		const K& q, nat n1, nat n2)
{
  if (n1 == 0 || n2 == 0) return;
  INLINE_ARRAY (C, t2, n2);
  jet_clear (dest, n1+n2-1);
  jet_copy (t2, s2, n2);
  while (n1 != 0) {
    jet_mul_add (dest, *s1, t2, n2);
    jet_coeff_q_difference (t2, q, n2);
    dest++; s1++; n1--;
  }
  DELETE_ARRAY (C, t2, n2);
}

template<typename C, typename K, const K& q>
class multiplier<C,skew<q_operator<K,q> > > {
public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    mul_q_operator (dest, s1, s2, q, n1, n2);
  }
};

template<typename C, typename K, const K& q>
class adjoint_helper<C,q_operator<K,q> > {
public:
  static void indeterminate (C* dest) {
    dest[0]= C(0);
    dest[1]= C(1);
  }
  static void mul (C* dest, const C* s1, const C* s2,
		   nat n1, nat n2, bool adj) {
    if (adj) jet_mul<C,skew<q_operator<K,inv_q<K,q>::val> > >
	       (dest, s1, s2, n1, n2);
    else jet_mul<C,skew<q_operator<K,q> > > (dest, s1, s2, n1, n2);
  }
};

template<typename C, typename K, const K& q> generic
var (const polynomial<C,dense<skew<q_operator<K,q> > > >& P) {
  (void) P;
  return as_generic (MML_ACCESS) (generic ("Q"), as_generic (q));
}

template<typename C, typename K, const K& q>
polynomial<C,dense<skew<q_operator<K,inv_q<K,q>::val> > > >
adjoint (const polynomial<C,dense<skew<q_operator<K,q> > > >& P) {
  nat n= dim (P);
  C* r= new C[n];
  if (n>1) jet_dicho_adjoint<C,q_operator<K,q> > (r, seg (P), n, false);
  return polynomial<C,dense<skew<q_operator<K,inv_q<K,q>::val> > > > (r, n);
}

} // namespace mml
#endif //__MML__SKEW_EXTRA__HPP
