
/******************************************************************************
* MODULE     : symbolic_arithmetic.hpp
* DESCRIPTION: Arithmeric operations on symbolic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_ARITHMETIC_HPP
#define __SYMBOLIC_ARITHMETIC_HPP
#include "symbolic.hpp"
namespace mml {
#define TMPL template<typename C,typename V>
#define Symbolic symbolic<C,V>
#define Vector vector<Symbolic >
#define Table table<Symbolic,Symbolic >
#define Iterator iterator<Symbolic >

/******************************************************************************
* Prototypes
******************************************************************************/

TMPL Symbolic operator - (const Symbolic& x);
TMPL Symbolic operator + (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic operator - (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic operator * (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic operator / (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic operator % (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic pow (const Symbolic& x1, const Symbolic& x2);
TMPL Symbolic pow (const Symbolic& x, const int& i);
TMPL Symbolic pow (const int& i, const Symbolic& x);
TMPL Symbolic sum (const Vector& v);
TMPL Symbolic product (const Vector& v);

/******************************************************************************
* Addition and subtraction
******************************************************************************/

TMPL Table
get_sum (const Symbolic& x) {
  if (is_sum (x)) return as_sum (x);
  Table t (0);
  if (is_scalar (x)) t(1)= x;
  else t(x)= 1;
  return t;
}

TMPL Symbolic
make_sum (const Table& t) {
  Symbolic sc= 0;
  Table r (0);
  for (Iterator it= entries (t); busy (it); ++it) {
    Symbolic x= *it;
    if (t[x] == 0);
    else if (is_scalar (x))
      sc= as_scalar (sc) + as_scalar (x) * as_scalar (t[x]);
    else r(x)= t[x];
  }
  if (sc != 0) r(1)= sc;

  if (N(r) == 0) return 0;
  if (N(r) == 1) {
    Symbolic x= *entries (r);
    if (x == 1) return r[x];
    if (r[x] == 1) return x;
  }
  return symbolic_sum (r);
}

TMPL Symbolic
neg (const Symbolic& x1) {
  if (is_scalar (x1)) return -as_scalar (x1);
  if (x1 == 0) return x1;
  return make_sum (-get_sum (x1));
}

TMPL Symbolic
add (const Symbolic& x1, const Symbolic& x2) {
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) + as_scalar (x2);
  if (x1 == 0) return x2;
  if (x2 == 0) return x1;
  return make_sum (get_sum (x1) + get_sum (x2));
}

TMPL Symbolic
sub (const Symbolic& x1, const Symbolic& x2) {
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) - as_scalar (x2);
  if (x2 == 0) return x1;
  return make_sum (get_sum (x1) - get_sum (x2));
}

/******************************************************************************
* Multiplication
******************************************************************************/

TMPL Table
get_product (const Symbolic& x) {
  if (is_product (x)) return as_product (x);
  if (is_sum (x)) {
    Table sum= as_sum (x);
    if (N (sum) == 1) {
      Symbolic y= *entries (sum);
      Table t (0);
      if (is_product (y)) t= copy (as_product (y));
      else t(y)= 1;
      t(sum[y])= 1;
      return t;
    }
  }
  Table t (0);
  if (x != 1) t(x)= 1;
  return t;
}

TMPL Symbolic
make_product (const Table& t) {
  Symbolic sc= 1;
  Table r (0);
  for (Iterator it= entries (t); busy (it); ++it) {
    Symbolic x= *it;
    if (t[x] == 0);
    else if (is_scalar (x)) {
      if (t[x] == 1) sc= as_scalar (sc) * as_scalar (x);
      else if (t[x] == -1) sc= as_scalar (sc) / as_scalar (x);
      else r(x)= t[x];
    }
    else r(x)= t[x];
  }

  if (N(r) == 0) return sc;
  if (N(r) == 1) {
    Symbolic x= *entries (t);
    if (t[x] == 1) {
      if (sc == 1) return x;
      Table u (0);
      u(x)= sc;
      return symbolic_sum (u);
    }
  }
  if (sc == 1) return symbolic_product (r);
  Table sum (0);
  sum (symbolic_product (r))= sc;
  return symbolic_sum (sum);
}

TMPL void
decompose_nc (const Symbolic& x, Symbolic& sc, Symbolic& y) {
  if (is_scalar (x)) {
    sc= x;
    y = 1;
  }
  else if (is_sum (x) && N (as_sum (x)) == 1) {
    Table t= as_sum (x);
    y = *entries (t);
    sc= t[y];
  }
  else {
    sc= 1;
    y = x;
  }
}

TMPL Symbolic
mul_nc (const Symbolic& x1, const Symbolic& x2) {
  /* normalize */
  if (is_func (x1, MML_TIMES, 2))
    return mul_nc (x1[1], mul_nc (x1[2], x2));
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) * as_scalar (x2);

  /* write x2 = h * r, with possibility that r = 1 */
  Symbolic h= x2, r= 1;
  if (is_func (x2, MML_TIMES, 2)) {
    h= x2[1];
    r= x2[2];
  }
  if (is_scalar (x1) && is_scalar (h))
    return mul_nc (Symbolic (as_scalar (x1) * as_scalar (h)), r);

  /* write x1 = rad1 ^ pow1 and h = rad2 ^ pow2 */
  Symbolic rad1= x1, rad2= h;
  Symbolic pow1= 1, pow2= 1;
  if (is_func (x1, MML_POWER, 2)) {
    rad1= x1[1];
    pow1= x1[2];
  }
  if (is_func (h, MML_POWER, 2)) {
    rad2= h[1];
    pow2= h[2];
  }

  /* regroup powers if rad1 = rad2 */
  Symbolic op1= x1, op2= x2;
  if (rad1 == rad2)
    if (V::general_power_add ||
	(V::scalar_power_add && (is_scalar (pow1) || is_scalar (pow2))))
      {
	op1= pow (rad1, pow1 + pow2);
	op2= r;
      }
  if (op1 == 1) return op2;
  if (op2 == 1) return op1;
  return Symbolic (MML_TIMES) (op1, op2);
}

TMPL Symbolic
mul_term (const Symbolic& x1, const Symbolic& x2) {
  if (V::general_commute)
    return make_product (get_product (x1) + get_product (x2));
  else {
    Symbolic sc1, sc2, y1, y2, sc, y;
    decompose_nc (x1, sc1, y1);
    decompose_nc (x2, sc2, y2);
    if (!V::scalar_commute && sc2 != 1 && sc2 != -1) {
      y2 = Symbolic (MML_TIMES) (sc2, y2);
      sc2= 1;
    }
    sc= as_scalar (sc1) * as_scalar (sc2);
    y = mul_nc (y1, y2);
    if (is_scalar (y)) return as_scalar (sc) * as_scalar (y);
    if (sc == 0) return 0;
    if (sc == 1) return y;
    Table t (0);
    t(y)= sc;
    return symbolic_sum (t);
  }
}

TMPL Table
mul (const Table& t1, const Table& t2) {
  Table r (0);
  for (Iterator it1= entries (t1); busy (it1); ++it1)
    for (Iterator it2= entries (t2); busy (it2); ++it2) {
      if (!V::scalar_commute) {
	Symbolic x= mul_term (mul_term (*it1, t2[*it2]), *it2);
	r(x) += t1[*it1];
      }
      else {
	Symbolic x= mul_term (*it1, *it2);
	r(x) += t1[*it1] * t2[*it2];
      }
    }
  return r;
}

TMPL Symbolic
mul (const Symbolic& x1, const Symbolic& x2) {
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) * as_scalar (x2);
  if (x1 == 0) return 0;
  if (x2 == 0) return 0;
  if (x1 == 1) return x2;
  if (x2 == 1) return x1;
  if (V::scalar_expand) {
    if (is_scalar (x1)) return make_sum (x1 * get_sum (x2));
    if (V::scalar_commute && is_scalar (x2))
      return make_sum (x2 * get_sum (x1));
  }
  if (V::general_expand) return make_sum (mul (get_sum (x1), get_sum (x2)));
  else return mul_term (x1, x2);
}

/******************************************************************************
* Division
******************************************************************************/

TMPL Symbolic
reorder_nc (const Symbolic& x) {
  if (!is_func (x, MML_TIMES, 2)) return x;
  else {
    Symbolic y= x[1], r= x[2];
    while (is_func (y, MML_TIMES, 2)) {
      r= Symbolic (MML_TIMES) (y[2], r);
      y= y[1];
    }
    return Symbolic (MML_TIMES) (y, r);
  }
}

TMPL Symbolic
inv_nc (const Symbolic& x) {
  /* normalize */
  if (is_scalar (x)) return C(1) / as_scalar (x);

  /* write x = h * r, with possibility that r = 1 */
  Symbolic h= x, r= 1;
  if (is_func (x, MML_TIMES, 2)) {
    h= x[1];
    r= x[2];
  }

  /* write h = rad_h ^ pow_h */
  Symbolic rad_h= h;
  Symbolic pow_h= 1;
  if (is_func (h, MML_POWER, 2)) {
    rad_h= h[1];
    pow_h= h[2];
  }

  /* regroup powers if rad1 = rad2 */
  Symbolic op= pow (rad_h, -pow_h);
  if (r == 1) return op;
  return Symbolic (MML_TIMES) (inv_nc (r), op);
}

TMPL Symbolic
div_term (const Symbolic& x1, const Symbolic& x2) {
  if (V::general_commute)
    return make_product (get_product (x1) - get_product (x2));
  else {
    Symbolic sc1, sc2, y1, y2, sc, y;
    decompose_nc (x1, sc1, y1);
    decompose_nc (x2, sc2, y2);
    if (!V::scalar_commute && sc2 != 1 && sc2 != -1) {
      y2 = Symbolic (MML_TIMES) (sc2, y2);
      sc2= 1;
    }
    sc= as_scalar (sc1) / as_scalar (sc2);
    y = mul_nc (y1, reorder_nc (inv_nc (y2)));
    if (is_scalar (y)) return as_scalar (sc) * as_scalar (y);
    if (sc == 0) return 0;
    if (sc == 1) return y;
    Table t (0);
    t(y)= sc;
    return symbolic_sum (t);
  }
}

TMPL Table
div (const Table& t, const Symbolic& sc, const Symbolic& x) {
  Table r (0);
  for (Iterator it= entries (t); busy (it); ++it) {
    Symbolic y= div_term (*it, x);
    r(y)= div_term (t[*it], sc);
  }
  return r;
}

TMPL Symbolic
div (const Symbolic& x1, const Symbolic& x2) {
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) / as_scalar (x2);
  if (x1 == 0) return 0;
  if (x2 == 1) return x1;
  if (x2 == 0) return 0;
  if (V::scalar_expand && V::scalar_commute && is_scalar (x2))
    return make_sum (get_sum (x1) / x2);
  if (V::general_expand) {
    Table t= get_sum (x2);
    if (V::scalar_commute && N(t) == 1) {
      Symbolic y = *entries (t);
      Symbolic sc= t[y];
      return make_sum (div (get_sum (x1), sc, y));
    }
    else return make_sum (div (get_sum (x1), Symbolic (1), x2));
  }
  else return div_term (x1, x2);
}

TMPL Symbolic
mod (const Symbolic& x1, const Symbolic& x2) {
  if (is_scalar (x1) && is_scalar (x2))
    return as_scalar (x1) % as_scalar (x2);
  if (x2 == 0) return x1;
  return 0;
}

/******************************************************************************
* Powers
******************************************************************************/

TMPL Symbolic
power (const Symbolic& x1, const Symbolic& x2) {
  if (x1 == 0) return x1;
  if (x1 == 1) return x1;
  if (x2 == 0) return 1;
  if (x2 == 1) return x1;
  if (x2 == -1 && is_scalar (x1)) return 1 / x1;
  if (V::general_commute) {
    Table t= get_product (x1);
    if (V::general_power_expand || (V::scalar_power_expand && is_scalar (x2)))
      return make_product (x2 * t);
    if (N(t) == 1)
      if (V::general_power_multiply ||
	  (V::scalar_power_multiply && is_scalar (x2)))
	return make_product (x2 * t);
    Table r (0);
    r(x1)= x2;
    return symbolic_product (r);
  }
  if (is_func (x1, MML_POWER, 2))
    if (V::general_power_multiply ||
	(V::scalar_power_multiply && is_scalar (x2)))
      return power (x1[1], x1[2] * x2);
  return Symbolic (MML_POWER) (x1, x2);
}

/******************************************************************************
* Big sums and products
******************************************************************************/

TMPL Symbolic
big_sum (const Vector& v) {
  Table r (0);
  nat i, n= dim (v);
  for (i=0; i<n; i++) {
    Table t= get_sum (v[i]);
    for (Iterator it= entries (t); busy (it); ++it)
      r(*it) += t[*it];
  }
  return make_sum (r);
}

TMPL Symbolic
big_product (const Vector& v) {
  if (V::general_commute && !V::general_expand) {
    Table r (0);
    nat i, n= dim (v);
    for (i=0; i<n; i++) {
      Table t= get_product (v[i]);
      for (Iterator it= entries (t); busy (it); ++it)
	r(*it) += t[*it];
    }
    return make_product (r);
  }
  else {
    Symbolic r= 1;
    nat i, n= dim (v);
    for (i=n-1; i!=((nat) -1); i--)
      r= v[i] * r;
    return r;
  }
}

/******************************************************************************
* Public interface
******************************************************************************/

TMPL Symbolic operator - (const Symbolic& x1) {
  return neg (x1); }
TMPL Symbolic operator + (const Symbolic& x1, const Symbolic& x2) {
  return add (x1, x2); }
TMPL Symbolic operator - (const Symbolic& x1, const Symbolic& x2) {
  return sub (x1, x2); }
TMPL Symbolic operator * (const Symbolic& x1, const Symbolic& x2) {
  return mul (x1, x2); }
TMPL Symbolic operator / (const Symbolic& x1, const Symbolic& x2) {
  return div (x1, x2); }
TMPL Symbolic operator % (const Symbolic& x1, const Symbolic& x2) {
  return mod (x1, x2); }
TMPL Symbolic pow (const Symbolic& x1, const Symbolic& x2) {
  return power (x1, x2); }
TMPL Symbolic pow (const Symbolic& x, const int& i) {
  return power (x, Symbolic (i)); }
TMPL Symbolic pow (const int& i, const Symbolic& x) {
  return power (Symbolic (i), x); }
TMPL Symbolic sum (const Vector& v) {
  return big_sum (v); }
TMPL Symbolic product (const Vector& v) {
  return big_product (v); }

ARITH_INT_SUGAR(TMPL,Symbolic,Symbolic);

#undef TMPL
#undef Symbolic
#undef Vector
#undef Table
#undef Iterator
} // namespace mml
#endif // __SYMBOLIC_ARITHMETIC_HPP
