
/******************************************************************************
* MODULE     : generic_pretty.cpp
* DESCRIPTION: Pretty printer for generic expressions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "generic_internal.hpp"
#include "list.hpp"
namespace mml {

/******************************************************************************
* The pretty printer class
******************************************************************************/

class pretty_printer {
private:
  ostream&      out;
  list<generic> gl;

private:
  void pp_verb (const string& s);
  void pp_op (const generic& op, bool infix= true);
  void pp_keyw (bool before, const string& s, bool after);

  void pp_E0s (const generic& g, nat start);
  void pp_Vtuple (const generic& g, nat start);

  void pp_E1 (const generic& g);
  void pp_Lflux (const generic& g);
  void pp_Rflux (const generic& g);

  void pp_F0 (const generic& g);
  void pp_Ors (const generic& g);
  void pp_Xors (const generic& g);
  void pp_F1 (const generic& g);
  void pp_Ands (const generic& g);
  void pp_F2 (const generic& g);
  void pp_F3 (const generic& g);

  void pp_T0 (const generic& g);
  void pp_T1 (const generic& g);
  void pp_T2 (const generic& g);
  void pp_T3 (const generic& g);
  void pp_T4 (const generic& g);
  void pp_T5 (const generic& g);

  void pp_R0 (const generic& g);
  void pp_R1 (const generic& g);
  void pp_R2 (const generic& g);
  void pp_R3 (const generic& g);

public:
  inline pretty_printer (ostream& out2): out (out2), gl () {}
  inline ~pretty_printer () {}

  void pp_E0 (const generic& g);
};

/******************************************************************************
* Helper routines
******************************************************************************/

bool
starts (const string& s1, const string& s2) {
  return N(s1) >= N(s2) && s1 (0, N(s2)) == s2;
}

void
pretty_printer::pp_verb (const string& s) {
  out << s;
}

void
pretty_printer::pp_op (const generic& op, bool infix) {
  if (infix) out << " ";
  out << as_string (op);
  if (infix) out << " ";
}

void
pretty_printer::pp_keyw (bool before, const string& s, bool after) {
  if (before) out << " ";
  out << s;
  if (after) out << " ";
}

/******************************************************************************
* Tuples and vertical tuples
******************************************************************************/

void
pretty_printer::pp_E0s (const generic& g, nat start) {
  nat n= N (g);
  for (nat i=start; i<n; i++) {
    if (i != start) pp_verb (", ");
    pp_E0 (g[i]);
  }
}

void
pretty_printer::pp_Vtuple (const generic& g, nat start) {
  nat n= N (g);
  for (nat i=start; i<n; i++) {
    if (i != start) pp_verb (";\n");
    if (is_func (g[i], MML_TUPLE)) pp_E0s (g[i], 1);
    else pp_E0 (g[i]);
  }
}

/******************************************************************************
* Expressions
******************************************************************************/

void
pretty_printer::pp_E0 (const generic& g) {
  pp_E1 (g);
}

void
pretty_printer::pp_E1 (const generic& g) {
  if (is_func (g, MML_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op (g[0]);
    pp_F0 (g[2]);
  }
  else if (is_func (g, MML_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (g[0]);
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

void
pretty_printer::pp_Lflux (const generic& g) {
  if (is_func (g, MML_LESSLESS, 2)) {
    pp_Lflux (g[1]);
    pp_op (g[0]);
    pp_F0 (g[2]);
  }
  else pp_F0 (g);
}

void
pretty_printer::pp_Rflux (const generic& g) {
  if (is_func (g, MML_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (g[0]);
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

/******************************************************************************
* Formulas
******************************************************************************/

void
pretty_printer::pp_F0 (const generic& g) {
  if (is_func (g, MML_IMPLIES, 2) ||
      is_func (g, MML_EQUIV, 2)) {
    pp_F1 (g[1]);
    pp_op (g[0]);
    pp_F1 (g[2]);
  }
  else pp_F1 (g);
}

void
pretty_printer::pp_F1 (const generic& g) {
  if (is_func (g, MML_OR, 2)) {
    pp_Xors (g[1]);
    pp_op (g[0]);
    pp_Xors (g[2]);
  }
  else if (is_func (g, MML_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (g[0]);
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer::pp_Ors (const generic& g) {
  if (is_func (g, MML_OR, 2)) {
    pp_Ors (g[1]);
    pp_op (g[0]);
    pp_Ors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer::pp_Xors (const generic& g) {
  if (is_func (g, MML_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (g[0]);
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer::pp_F2 (const generic& g) {
  if (is_func (g, MML_AND, 2)) {
    pp_Ands (g[1]);
    pp_op (g[0]);
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pretty_printer::pp_Ands (const generic& g) {
  if (is_func (g, MML_AND, 2)) {
    pp_Ands (g[1]);
    pp_op (g[0]);
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pretty_printer::pp_F3 (const generic& g) {
  if (is_func (g, MML_EQUAL, 2) ||
      is_func (g, MML_UNEQUAL, 2) ||
      is_func (g, MML_LESS, 2) ||
      is_func (g, MML_LESSEQ, 2) ||
      is_func (g, MML_GTR, 2) ||
      is_func (g, MML_GTREQ, 2)) {
    pp_T0 (g[1]);
    pp_op (g[0]);
    pp_T0 (g[2]);
  }
  else pp_T0 (g);
}

/******************************************************************************
* Terms
******************************************************************************/

void
pretty_printer::pp_T0 (const generic& g) {
  pp_T1 (g);
}

void
pretty_printer::pp_T1 (const generic& g) {
  if (is_func (g, MML_FUNCTION, 2)) {
    pp_T2 (g[1]);
    pp_op (g[0]);
    pp_T2 (g[2]);
  }
  else pp_T2 (g);
}

void
pretty_printer::pp_T2 (const generic& g) {
  pp_T3 (g);
}

void
pretty_printer::pp_T3 (const generic& g) {
  if (is_func (g, MML_PLUS, 2)) {
    pp_T3 (g[1]);
    pp_op (g[0]);
    pp_T3 (g[2]);
  }
  else if (is_func (g, MML_MINUS, 2)) {
    pp_T3 (g[1]);
    pp_op (g[0]);
    pp_T4 (g[2]);
  }
  else pp_T4 (g);
}

void
pretty_printer::pp_T4 (const generic& g) {
  if (is_func (g, MML_TIMES, 2) ||
      is_func (g, MML_COMPOSE, 2)) {
    pp_T4 (g[1]);
    pp_op (g[0]);
    pp_T4 (g[2]);
  }
  else if (is_func (g, MML_OVER, 2)) {
    pp_T4 (g[1]);
    pp_op (g[0]);
    pp_T5 (g[2]);
  }
  else if (is_func (g, MML_DIV, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "div", true);
    pp_T5 (g[2]);
  }
  else if (is_func (g, MML_MOD, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "mod", true);
    pp_T5 (g[2]);
  }
  else pp_T5 (g);
}

void
pretty_printer::pp_T5 (const generic& g) {
  if (is_func (g, MML_NOT, 1) ||
      is_func (g, MML_MINUS, 1)) {
    pp_op (g[0], false);
    pp_R0 (g[1]);
  }
  else if (is_atom (g) && starts (as_string (g), "-"))
    pp_verb (as_string (g));
  else pp_R0 (g);
}

/******************************************************************************
* Radicals
******************************************************************************/

void
pretty_printer::pp_R0 (const generic& g) {
  pp_R1 (g);
}

void
pretty_printer::pp_R1 (const generic& g) {
  if (is_func (g, MML_POWER, 2)) {
    pp_R2 (g[1]);
    pp_op (g[0], false); // suppress spaces
    pp_R2 (g[2]);
  }
  else pp_R2 (g);
}

static bool
is_applicable (const generic& g) {
  if (is_atom (g)) {
    string s= as_string (g);
    if (N(s)==0) return false;
    for (nat i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]<'_') || (s[i]>'_')))
        return false;
  }
  return true;
}

void
pretty_printer::pp_R2 (const generic& g) {
  if (is_func (g, MML_FACTORIAL, 1)) {
    pp_R2 (g[1]);
    pp_verb ("!");
  }
  else if (is_func (g, MML_PRIME, 1)) {
    pp_R2 (g[1]);
    pp_op (g[0], false);
  }
  else if (is_func (g, MML_DOT, 2)) {
    pp_R2 (g[1]);
    pp_verb (".");
    pp_R3 (g[2]);
  }
  else if (is_func (g, MML_ACCESS) && N (g) > 2) {
    pp_R2 (g[1]);
    pp_verb ("[");
    if (N (g) == 3 && is_func (g[2], MML_VTUPLE))
      pp_Vtuple (g[2], 1);
    else pp_E0s (g, 2);
    pp_verb ("]");
  }
  else if ((!is_atom (g) && is_applicable (g[0])) ||
	   (!nil (gl)) && (car (gl) == g)) {
    pp_R2  (g[0]);
    pp_verb (" (");
    pp_E0s (g, 1);
    pp_verb (")");
  }
  else pp_R3 (g);
}

void
pretty_printer::pp_R3 (const generic& g) {
  if (is_atom (g) && !starts (as_string (g), "-")) {
    if (g == MML_E) pp_verb ("e");
    else if (g == MML_PI) pp_verb ("pi");
    else if (g == MML_PARTIAL) pp_verb ("d");
    else if (g == MML_DELTA) pp_verb ("d");
    else pp_verb (as_string (g));
  }
  else if (is_func (g, MML_SQTUPLE)) {
    pp_verb ("[");
    if (N(g) == 2 && is_func (g[1], MML_VTUPLE))
      pp_Vtuple (g[1], 1);
    else pp_E0s (g, 1);
    pp_verb ("]");
  }
  else if (is_func (g, MML_TUPLE)) {
    pp_verb ("(");
    pp_E0s (g, 1);
    pp_verb (")");
  }
  else if (is_func (g, MML_VTUPLE)) {
    pp_verb ("(");
    pp_Vtuple (g, 1);
    pp_verb (")");
  }
  else {
    pp_verb ("(");
    gl= cons (g, gl);
    pp_E0 (g);
    gl= cdr (gl);
    pp_verb (")");
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
pretty_print (ostream& out, const generic& g) {
  pretty_printer pp (out);
  pp.pp_E0 (g);
}

} // namespace mml
